/*
 * #%L
 * EUGene :: EUGene
 * 
 * $Id: AbstractMetaTransformer.java 906 2010-05-16 12:19:55Z tchemit $
 * $HeadURL: http://svn.nuiton.org/svn/eugene/tags/eugene-2.1/eugene/src/main/java/org/nuiton/eugene/AbstractMetaTransformer.java $
 * %%
 * Copyright (C) 2004 - 2010 CodeLutin
 * %%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation, either version 3 of the 
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Lesser Public License for more details.
 * 
 * You should have received a copy of the GNU General Lesser Public 
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/lgpl-3.0.html>.
 * #L%
 */

package org.nuiton.eugene;

import org.nuiton.eugene.models.Model;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Abstract meta transformer which contains some templates to apply to an
 * incoming model.
 *
 * Using the {@link #getExcludeTemplates()} to restrict use of some templates.
 * 
 * Created: 20 déc. 2009
 *
 * @param <M> type of  Model
 * 
 * @author Tony Chemit <chemit@codelutin.com>
 * @since 2.0.0
 */
public abstract class AbstractMetaTransformer<M extends Model> extends AbstractGenerator<M> {

    private static final Log log =
            LogFactory.getLog(AbstractMetaTransformer.class);

    private final Class<? extends Template<M>>[] transformers;

    public AbstractMetaTransformer(
            Class<? extends Template<M>>... transformers) {
        this.transformers = transformers;
    }

    protected abstract boolean validateModel(M model);

    @Override
    public void applyTemplate(M model, File destDir) throws IOException {

        boolean isValid = validateModel(model);
        if (!isValid) {

            //On n'empeche pas la génération si le modèle n'est pas valide

            return;
        }

        List<Template<M>> templates = getTemplates();

        for (Template<M> template : templates) {
            // log
            if (log.isDebugEnabled()) {
                log.debug("call template : " +
                          template.getClass().getSimpleName());
            }
            template.applyTemplate(model, destDir);
        }
    }

    protected List<Template<M>> getTemplates() {
        List<Template<M>> result = new ArrayList<Template<M>>();
        for (Class<? extends Template<M>> generatorClass : getTransformers()) {
            Template<M> generator;
            if (getExcludeTemplates().contains(generatorClass.getName())) {
                // exclude generator
                if (log.isInfoEnabled()) {
                    log.info("exclude generator " + generatorClass);
                }
                continue;
            }
            try {
                generator = generatorClass.newInstance();

                // init template

                generator.setConfiguration(getConfiguration());
                
                if (generator instanceof AbstractGenerator) {
                    ((AbstractGenerator<M>) generator).setParent(this);
                }
//                if (generator instanceof Transformer) {
//                    generator.setProperties(properties);
//                }

                // will use it
                result.add(generator);
            } catch (Exception e) {
                // should never happens
                if (log.isErrorEnabled()) {
                    log.error("An error occurs when generating persistence", e);
                }
                throw new RuntimeException(e);
            }
        }
        return result;
    }

    public Class<? extends Template<M>>[] getTransformers() {
        return transformers;
    }
}
