/*
 * #%L
 * EUGene :: EUGene
 * 
 * $Id: DefaultChainedWriterEngine.java 863 2010-04-15 14:22:49Z tchemit $
 * $HeadURL: http://svn.nuiton.org/svn/eugene/tags/eugene-2.1/eugene/src/main/java/org/nuiton/eugene/writer/DefaultChainedWriterEngine.java $
 * %%
 * Copyright (C) 2004 - 2010 CodeLutin
 * %%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation, either version 3 of the 
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Lesser Public License for more details.
 * 
 * You should have received a copy of the GNU General Lesser Public 
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/lgpl-3.0.html>.
 * #L%
 */

package org.nuiton.eugene.writer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.File;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Default implementation of the {@link ChainedWriterEngine}.
 * <p/>
 * Created: 17 déc. 2009
 *
 * @author Tony Chemit <chemit@codelutin.com> Copyright Code Lutin
 * @version $Revision: 863 $
 *          <p/>
 *          Mise a jour: $Date: 2010-04-15 16:22:49 +0200 (jeu., 15 avril 2010) $ par :
 *          * @plexus.component role="org.nuiton.eugene.writer.ChainedWriterEngine" role-hint="default"
 * @since 2.0.0
 */
public class DefaultChainedWriterEngine implements ChainedWriterEngine {

    /**
     * Logger
     */
    private static final Log log =
            LogFactory.getLog(DefaultChainedWriterEngine.class);
    /**
     * shared configuration
     */
    protected ChainedFileWriterConfiguration configuration;
    /**
     * available writers corresponding to the given configuration
     */
    protected Set<ChainedFileWriter> availableWriters;
    /**
     * selected writers obtain while the register phase.
     */
    protected List<ChainedFileWriter> selectedWriters;

    @Override
    public ChainedFileWriterConfiguration getConfiguration() {
        return configuration;
    }

    @Override
    public void setConfiguration(ChainedFileWriterConfiguration configuration) {
        this.configuration = configuration;
    }

    @Override
    public List<ChainedFileWriter> getSelectedWriters() {
        if (selectedWriters == null) {
            checkConfiguration("getSelectedWriters");
            selectedWriters = new ArrayList<ChainedFileWriter>();
        }
        return selectedWriters;
    }

    @Override
    public boolean containsWriter(String inputProtocol) {
        for (ChainedFileWriter w : getSelectedWriters()) {
            if (inputProtocol.equals(w.getInputProtocol())) {
                return true;
            }
        }
        return false;
    }

    @Override
    public Set<ChainedFileWriter> getAvailableWriters() {
        if (availableWriters == null) {
            checkConfiguration("getAvailableWriters");
            availableWriters = filterWriterForModelType(
                    getConfiguration().getWriters(),
                    getConfiguration().getModelType()
            );
        }
        return availableWriters;
    }

    @Override
    public void registerInclude(String include) {
        checkConfiguration("registerInclude");

        List<ChainedFileWriter> selectedWriters = getSelectedWriters();

        if (log.isDebugEnabled()) {
            log.debug(">>>>>>>>>>>>>>>>> [" + include + "]");
            log.debug("Will register include " + "[" + include + "]");
            log.debug("actual selected writers : " + selectedWriters);
        }
        ModelFileWriterEntryType selectedType = null;
        Matcher matcher = null;

        // obtain the good type of entry and the corresponding matcher

        for (ModelFileWriterEntryType type :
                ModelFileWriterEntryType.values()) {
            matcher = type.getMatcher(include);
            if (matcher != null) {
                // get a matcher
                selectedType = type;
                break;
            }
        }

        if (selectedType == null) {
            // no writer
            throw new IllegalArgumentException("could not find a writer for " +
                                               "include pattern : " + include);
        }

        // obtain the writer
        ChainedFileWriter writer =
                selectedType.getWriter(this, include, matcher);

        // create the new entry

        ChainedFileWriterEntry writerEntry =
                selectedType.newEntry(this, include, matcher, writer);

        // register the new entry

        writer.addEntry(writerEntry);
        if (!selectedWriters.contains(writer)) {
            // register the writer as to be used
            if (log.isDebugEnabled()) {
                log.debug("[" + include + "] Associated with writer " + writer);
            }
            selectedWriters.add(writer);
        }

        ChainedFileWriterConfiguration configuration = getConfiguration();

        String modelType = configuration.getModelType();

        String outpoutProtocol = writer.getOutputProtocol(modelType);

        if (outpoutProtocol == null) {
            // nothing more to do
            if (log.isDebugEnabled()) {
                log.debug("<<<<<<<<<<<<<<<<< [" + include + "]");
            }
            return;
        }

        // the writer need the includes of another writer

        if (log.isDebugEnabled()) {
            log.debug("[" + include + "]" + " writer " +
                      writer.getClass().getSimpleName() +
                      " require a next writer of protocol " + outpoutProtocol);
        }

        ChainedFileWriter nextWriter =
                ((AbstractChainedFileWriter) writer).getNextWriter();
        if (nextWriter == null) {
            // the next writer was not initialize, just have to add new entry

            nextWriter = getWriterForInputProtocol(
                    getAvailableWriters(),
                    outpoutProtocol,
                    modelType
            );

            if (nextWriter == null) {
                throw new IllegalArgumentException(
                        "could not find a writer for protocole " +
                        outpoutProtocol + " on model " + modelType);
            }
            // chain writer
            ((AbstractChainedFileWriter) writer).setNextWriter(nextWriter);
        }

        // build the next include to do

        String basedirpath = configuration.getBasedir().getAbsolutePath();

        String outputpath = writer.getOutputDirectory(
                configuration.getOutputDirectory(),
                configuration.isTestPhase()).getAbsolutePath();

        String path = outputpath.substring(basedirpath.length() + 1);

        String newInclude = outpoutProtocol + ":" + path + ":" +
                            nextWriter.getDefaultIncludes();

        registerInclude(newInclude);

        if (log.isDebugEnabled()) {
            log.debug("<<<<<<<<<<<<<<<<< [" + include + "]");
        }
    }

    @Override
    public void clear() {
        if (selectedWriters != null) {
            selectedWriters.clear();
            selectedWriters = null;
        }
        if (availableWriters != null) {
            for (ChainedFileWriter writer : availableWriters) {
                writer.clear();
            }
            availableWriters.clear();
            availableWriters = null;
        }
        setConfiguration(null);
    }

    @Override
    public Set<ChainedFileWriter> filterWriterForModelType(
            Map<String, ChainedFileWriter> universe, String modelType) {
        Set<ChainedFileWriter> result = new HashSet<ChainedFileWriter>();
        for (ChainedFileWriter w : universe.values()) {
            if (w.acceptModel(modelType)) {
                if (log.isDebugEnabled()) {
                    log.debug("writer [" + w + "] accept model " + modelType);
                }
                result.add(w);
            }
        }
        return result;
    }

    @Override
    public ChainedFileWriter getWriterForInputProtocol(
            Set<ChainedFileWriter> universe,
            String inputProtocol,
            String modelType) {
        for (ChainedFileWriter writer : universe) {
            if (inputProtocol.equals(writer.getInputProtocol(modelType))) {
                return writer;
            }
        }
        return null;
    }

    @Override
    public ChainedFileWriter getWriterForInclude(
            Set<ChainedFileWriter> universe,
            String include,
            String modelType) {
        for (ChainedFileWriter w : universe) {
            if (w.acceptInclude(include)) {
                return w;
            }
        }
        return null;
    }

    protected void checkConfiguration(String method)
            throws IllegalStateException {
        if (getConfiguration() == null) {
            throw new IllegalStateException("can not acces to " + method +
                                            "before configuration is set!");
        }
    }

    /**
     *
     */
    public enum ModelFileWriterEntryType {

        ONLY_PROTOCOL_PATTERN(Pattern.compile("^([a-zA-Z]+)$")) {

            @Override
            public ChainedFileWriter getWriter(ChainedWriterEngine engine,
                                               String include,
                                               Matcher matcher) {
                ChainedFileWriterConfiguration configuration =
                        engine.getConfiguration();
                Set<ChainedFileWriter> universe = engine.getAvailableWriters();

                String protocol = matcher.group(1).toLowerCase();
                ChainedFileWriter writer = engine.getWriterForInputProtocol(
                        universe,
                        protocol,
                        configuration.getModelType()
                );
                if (writer == null) {
                    throw new IllegalArgumentException(
                            "could not find the writer named '" + protocol +
                            "', use one of " + universe);
                }
                if (log.isDebugEnabled()) {
                    log.debug("[" + include + "] " + "writer = (" + writer +
                              ")");
                }
                return writer;
            }

            @Override
            public ChainedFileWriterEntry newEntry(
                    ChainedWriterEngine engine,
                    String include,
                    Matcher matcher,
                    ChainedFileWriter writer) {
                ChainedFileWriterConfiguration configuration =
                        engine.getConfiguration();
                if (log.isDebugEnabled()) {
                    log.debug("[" + include + "] " + "detected pattern (" +
                              name() + ")");
                }

                ChainedFileWriterEntry writerEntry = new ChainedFileWriterEntry(
                        new File(configuration.getBasedir(),
                                 configuration.isTestPhase() ?
                                 writer.getDefaultTestInputDirectory() :
                                 writer.getDefaultInputDirectory()
                        ),
                        writer.getDefaultIncludes()
                );
                return writerEntry;
            }
        },

        NO_PROTOCOL_PATTERN(Pattern.compile("^([^:]+):([^:]+)$")) {
            @Override
            public ChainedFileWriter getWriter(ChainedWriterEngine engine,
                                               String include,
                                               Matcher matcher) {
                ChainedFileWriterConfiguration configuration =
                        engine.getConfiguration();
                Set<ChainedFileWriter> universe = engine.getAvailableWriters();

                String modelType = configuration.getModelType();
                // with no protocol pattern
                // pattern is discover from the includes

                // discover the writer from the given pattern
                ChainedFileWriter writer = engine.getWriterForInclude(
                        universe, include, modelType
                );
                if (writer == null) {
                    throw new IllegalArgumentException(
                            "could not find a writer for include " + include);
                }
                if (log.isDebugEnabled()) {
                    log.debug("[" + include + "] " + "writer = (" +
                              writer + ")");
                }
                return writer;
            }

            @Override
            public ChainedFileWriterEntry newEntry(ChainedWriterEngine engine,
                                                   String include,
                                                   Matcher matcher,
                                                   ChainedFileWriter writer) {
                ChainedFileWriterConfiguration configuration =
                        engine.getConfiguration();
                // with no protocol pattern
                // pattern is discover from the includes
                if (log.isDebugEnabled()) {
                    log.debug("[" + include + "] " + "detected pattern (" +
                              name() + ")");
                }

                String inputPath = matcher.group(1);
                String includes = matcher.group(2);

                ChainedFileWriterEntry writerEntry = new ChainedFileWriterEntry(
                        new File(configuration.getBasedir(), inputPath),
                        includes
                );
                return writerEntry;
            }
        },

        FULL_PATTERN(Pattern.compile("^(\\w+):([^:]+):([^:]+)$")) {
            @Override
            public ChainedFileWriter getWriter(ChainedWriterEngine engine,
                                               String include,
                                               Matcher matcher) {
                // with full pattern (protocol + directory + includes)
                // pattern is discover from the includes
                ChainedFileWriterConfiguration configuration =
                        engine.getConfiguration();
                Set<ChainedFileWriter> universe = engine.getAvailableWriters();

                String protocol = matcher.group(1).toLowerCase();
                ChainedFileWriter writer = engine.getWriterForInputProtocol(
                        universe, protocol, configuration.getModelType());

                if (writer == null) {
                    throw new IllegalArgumentException(
                            "could not find the writer named '" + protocol +
                            "', use one of " + universe);
                }
                if (log.isDebugEnabled()) {
                    log.debug("[" + include + "] " + "writer = (" + writer +
                              ")");
                }
                return writer;
            }

            @Override
            public ChainedFileWriterEntry newEntry(ChainedWriterEngine engine,
                                                   String include,
                                                   Matcher matcher,
                                                   ChainedFileWriter writer) {
                ChainedFileWriterConfiguration configuration =
                        engine.getConfiguration();
                // with full pattern (protocol + directory + includes)
                // pattern is discover from the includes
                if (log.isDebugEnabled()) {
                    log.debug("[" + include + "] " + "detected pattern (" +
                              name() + ")");
                }

                String inputPath = matcher.group(2);
                String includes = matcher.group(3);
                ChainedFileWriterEntry writerEntry = new ChainedFileWriterEntry(
                        new File(configuration.getBasedir(), inputPath),
                        includes
                );
                return writerEntry;
            }
        };

        private final Pattern pattern;

        ModelFileWriterEntryType(Pattern pattern) {
            this.pattern = pattern;
        }

        public Pattern getPattern() {
            return pattern;
        }

        public Matcher getMatcher(String include) {
            Matcher matcher = getPattern().matcher(include);
            return matcher.matches() ? matcher : null;
        }

        public abstract ChainedFileWriterEntry newEntry(
                ChainedWriterEngine engine,
                String include,
                Matcher matcher, ChainedFileWriter writer);

        public abstract ChainedFileWriter getWriter(
                ChainedWriterEngine engine,
                String include,
                Matcher matcher);
    }

}
